/* Flat polys with any number of vertexes. */



#include <string.h>
#include <math.h>

#include "../defs.h"

#include "../mem.h"

#include "../string.h"

#include "../texture.h"

#include "../lighttab.h"

#include "../asm/draw.inc"

#include "draw.h"



/* Draw a flat polygon with any number of vertexes (points).
pastXYZes have the array of polygon vertexes in
clockwise or anticlockwise order. nNumVertexes is the
number of vertexes in array. yCol is the polygon fill color.
Perform frame buffer clip. Medium optimized.
**********************************************************/
void CDraw::mFlatPolygon ( tagXY* pastXYZes,
                           INT nNumVertexes, BYTE yCol )
{
   // verify if have enough number of vertexes
   if ( nNumVertexes < 3 )
      ABORT ( MSG_ERROR_LOW_VERTEXES, __FILE__, __LINE__ )

   // seach top and bottom vertexes
   INT
      nYMin = pastXYZes [ 0 ].nY,   // y of the top vertex
      nYMax = pastXYZes [ 0 ].nY,   // y of the bottom vertex
      nPosMin = 0,       // the index of the top vertex (upper in screen)
      nPosMax = 0;       // the index of the bottom vertex (lower in screen)
   // verify vertexes
   for ( INT ncVert = 1; ncVert < nNumVertexes; ncVert ++ ) {

      // verify if array vertex is greater than the current bottom vertex
      if ( pastXYZes [ ncVert ].nY > nYMax ) {
         nYMax = pastXYZes [ ncVert ].nY;
         nPosMax = ncVert;
      }

      // verify if array vertex is less than the current top vertex
      if ( pastXYZes [ ncVert ].nY < nYMin ) {
         nYMin = pastXYZes [ ncVert ].nY;
         nPosMin = ncVert;
      }
   }

   // verify if top and bottom are the same, if yes abort polygon fill
   if ( nYMax == nYMin )
      return;

   // define color fill
   UINT
      uCol = yCol | ( yCol << 8 ) | ( yCol << 16 ) | ( yCol << 24 );

   // scan forward in array of vertexes starting on top and going down
   // With this you scan one side of the polygon.
   INT
      nPos1 = nPosMin,         // the first vertex to be tested
      nPos2 = nPos1;           // the next vertex (used in same test)
   while ( nPos1 != nPosMax ) {

      // define the next vertex position in vertex array
      nPos2 ++;
      // if no more vertexes go to array start
      if ( nPos2 == nNumVertexes )
         nPos2 = 0;

      // define polygon edges, the X for each scanline defined by the line
      // vertex 1 ----> vertex 2
      mScanFlatEdge ( m_pastForwardEdges, pastXYZes [ nPos1 ], pastXYZes [ nPos2 ] );

      // define current vertex as the next vertex
      nPos1 = nPos2;
   }

   // scan backward in array of vertexes starting on top and going down
   // With this you scan one side of the polygon.
   nPos1 = nPosMin;         // the first vertex to be tested
   nPos2 = nPos1;           // the next vertex (used in same test)
   while ( nPos1 != nPosMax ) {

      // define the previous vertex position in vertex array
      nPos2 --;
      // if no more vertexes go to array end
      if ( nPos2 == -1 )
         nPos2 = nNumVertexes - 1;

      // define polygon edges, the X for each scanline defined by the line
      // vertex 1 ----> vertex 2
      mScanFlatEdge ( m_pastBackwardEdges, pastXYZes [ nPos1 ], pastXYZes [ nPos2 ] );

      // define current vertex as the previous vertex
      nPos1 = nPos2;
   }

   // draw polygon
   // define scanline Y (if nYMin is less than 0, outside frame buff, use 0)
   INT
      nY = nYMin < 0 ? 0 : nYMin;
   // define memory position
   PTRVAL
      pLineAddress = (PTRVAL) m_poFrBuff->m_payMem + nY * m_nResX;

   // if X values for this scanline are swapped
   if ( m_pastBackwardEdges [ nY ].nX < m_pastForwardEdges [ nY ].nX )

      // while Y is upper than bottom vertex or bottom frame buffer
      while ( nY < nYMax && nY < m_nResY ) {
   
         mHorizontalLine ( m_pastBackwardEdges [ nY ],
                           m_pastForwardEdges [ nY ],
                           uCol, pLineAddress );

         // goto next scanline
         nY ++;
         pLineAddress += m_nResX;
      }
   else
      // while Y is upper than bottom vertex or bottom frame buffer
      while ( nY < nYMax && nY < m_nResY ) {
   
         mHorizontalLine ( m_pastForwardEdges [ nY ],
                           m_pastBackwardEdges [ nY ],
                           uCol, pLineAddress );
         // goto next scanline
         nY ++;
         pLineAddress += m_nResX;
      }
}



/* Used by CDraw::mSetClipEdgeBuffer to define X limits
for clip edge buffer and by CDraw::mFlatPolygon and to define
the X coordenates in each scanline for the polygon edges.
pastEdge - fill the x coordenates for each scanline;
stV1 and stV2 - vertex between scan must be done.
Always stV1 is less (topper) than stV2 in Y axis.
**********************************************************/
void CDraw::mScanFlatEdge ( tagPolyEdge* pastEdge,
                            tagXY& stV1, tagXY& stV2 )
{
   // verify if Y are in same line, if yes exit
   if ( stV1.nY == stV2.nY )
      return;

   INT
      // define where X start in first scanline, use fix point 16.16
      nX = stV1.nX << 16,
      // X increase rate
      nXInc = ( ( stV2.nX - stV1.nX ) << 16 ) /
                ( stV2.nY - stV1.nY );

   // do Y clip in frame buffer limits
   INT
      nYTop,
      nYBottom;
   // if y for vertex 1 is upper than frame buffer limits
   if ( stV1.nY < 0 ) {
      nYTop = 0;
      // if top Y has been cliped then is necessary update initial X value
      nX -= stV1.nY * nXInc;
   }
   else
      nYTop = stV1.nY;
   // if y for vertex 2 is lower than frame buffer limits
   if ( stV2.nY > m_nMaxY )
      nYBottom = m_nMaxY;
   else
      nYBottom = stV2.nY;

   // verify if need clip
   if ( nYBottom <= nYTop )
      return;

   // define X values in each scanline between top vertex and bottom vertex
   INT
      nYScan = nYTop;    // scanline in test
   do {

      pastEdge [ nYScan ].nX = nX >> 16; // X in frame buff, convert fix->int
      nX += nXInc;                       // new X in next scanline
      nYScan ++;                         // next scanline

   } while ( nYScan <= nYBottom );       // while not found bottom vertex
}



/* Draw a horizontal line.
stEdgeIni and stEdgeEnd have the start and end X for the line.
uCol - is an array with 4 bytes. Each elements must have the
color. Example: to draw a line with color 10 create the array:
char chCol [] = { 10, 10, 10, 10 };
pLineAddress is the position in frame buffer where the line
must be drawn.
High optimized. Clip only X values.
**********************************************************/
void CDraw::mHorizontalLine ( tagPolyEdge& stEdgeIni,
                              tagPolyEdge& stEdgeEnd, UINT uCol,
                              PTRVAL pLineAddress )
{
   // clip x values
   // verify start
   if ( stEdgeIni.nX < 0 )
      stEdgeIni.nX = 0;
   else
      // start can be in right (line outside frame buffer)
      if ( stEdgeIni.nX > m_nMaxX )
         return;
   // verify end
   if ( stEdgeEnd.nX > m_nMaxX )
      stEdgeEnd.nX = m_nMaxX;
   else
      // end can be in left (line outside frame buffer)
      if ( stEdgeEnd.nX < 0 )
         return;

   // draw horizontal line

   // if compile with asm code
   #ifdef COMP_ASM
   ASM_HLINE_FLAT_2 ( pLineAddress + stEdgeIni.nX , stEdgeEnd.nX - stEdgeIni.nX + 1, uCol );
   #else
   // compile C only (no asm code)
   memset ( (void*) ( pLineAddress + stEdgeIni.nX ), uCol, stEdgeEnd.nX - stEdgeIni.nX + 1 );
   #endif
}

