/* Implement keyboard functions to manage multiple keys presses without
repet delays.

This class is used for BIOS (DOS port) and X Window (Linux port) event handling.

Keyboard handler (only for BIOS) is a function that is called each time an
interruption 9d is genereted by the press of one or more keys. To access
keys status you can't use getch, scanf,... you need the functions
below, see tips to know how use getch, scanf,... with keyboard handler.

FEATURES:
- substitute BIOS keyboard function
- don't use keyboard buffer

TIPS:
- for BIOS: to use getch (), scanf () and other functions desinstall the keyboard
  handler with CKeyboard::mDesinstallKeyboardHandler (). After reinstall
  it using CKeyboard::mInstallKeyboardHandler ()
- The global array ayKEYS store key flags. If one key is pressed it's
  corresponding position in array is set. When is released it's position
  is reseted, more information below.

CAUTION:
- for BIOS: all keys must be released before desinstall the handler, or
  system will crach

BUGS:
- for BIOS: Print Screen and Control+Break are ignored while
  keyboard handler is on

TECHINFO:
(for BIOS)
- keyboard interruption: 9d
- port to read keys: 0x60
- port to signal interrupt captured: 0x20, value 0x20
- byte's values:
   If special key:
      give 224d from port 0x60 and
      next byte gived from port 0x60 is:
      If key pressed:
         key value        (array position: value + 128)
      If key released:
         key value + 128  (array position: value + 128)
   If not special key:
      If key pressed:
         key value        (array position: value)
      If key released:
         key value + 128  (array position: value)
*/



#ifdef TRG_DOSDJGPP
#include <dpmi.h>
#include <go32.h>
#include <string.h>
#include <dos.h>
#endif

#include "../defs.h"


// Data below must be global to enable handler and local functions access.
BOOL ayKEYS [ 256 ];    // key table: TRUE: given key is pressed else FALSE


#ifdef TRG_DOSDJGPP
BOOL bSPECIAL = FALSE;  // special key pressed
BOOL bINSTALL = FALSE;  // Signals handler is installed, if TRUE it can
                        // be desintalled else no.
_go32_dpmi_seginfo stOLD_KEYBOARD_HANDLER; // original keyboard handler
_go32_dpmi_seginfo stNEW_KEYBOARD_HANDLER; // this class keyboard handler
#endif



// the key codes for 3DFC convensions
#include "keycodes.h"



// include class declaration

// for DOS
#ifdef TRG_DOSDJGPP
#include "dosdjgpp/keyboard.h"
#endif

// for Linux X Window
#ifdef TRG_LINUXXWIN
#include "linxwin/keyboard.h"
#endif



/* constructor - if is not installed keyboard handler install it
 (only BIOS keyboard handling)
**********************************************************/
CKeyboard::CKeyboard ()
{
   #ifdef TRG_DOSDJGPP
   // verify if need installation
   if ( ! bINSTALL )
      mInstallKeyboardHandler ();
   #else
   // reset key table (all FALSE)
   memset ( (void*) ayKEYS, FALSE, sizeof ( ayKEYS ) );
   #endif
}



/* destructor - desinstall the keyboard handler if:
 it is installed and only if this instance has installed it
 (only BIOS keyboard handling)
**********************************************************/
CKeyboard::~CKeyboard ()
{
   #ifdef TRG_DOSDJGPP
   // verify if must try desinstall handler
   if ( m_bInstallHandler )
      mDesinstallKeyboardHandler ();
   #endif
}



// Functions implemented only on DOS port
#ifdef TRG_DOSDJGPP

/* Insplements the keyboard handler. Each time is generated an
interruption request this function is called.
SEE NewKeyboardHandler___ !!!!!
**********************************************************/
void NewKeyboardHandler ()
{
   // read key value
   BYTE
      yKey = inportb ( 0x60 );

   // verify if last call is not by an special key press
   if ( ! bSPECIAL )
      // detect special key press
      if ( yKey == 224 )
         bSPECIAL = TRUE;
      else
         // If not special key press, update table of key presses.
         // verify if key is beeing pressed (bits 7 == 0)
         if ( ( yKey & 128 ) == 0 )
            // if press set TRUE
            * ( ayKEYS + ( yKey & 127 ) ) = TRUE;
         else
            // if release (bit 7 == 1) set FALSE
            * ( ayKEYS + ( yKey & 127 ) ) = FALSE;
   else {
      // if special key had be pressed
      // verify if key is beeing pressed (bits 7 == 0)
      if ( ( yKey & 128 ) == 0 )
         * ( ayKEYS + 128 + ( yKey & 127 ) ) = TRUE;
      else
         // if release (bit 7 == 1) set FALSE
         * ( ayKEYS + 128 + ( yKey & 127 ) ) = FALSE;

      bSPECIAL = FALSE;
   }

   // signal insterruption end
   outportb ( 0x20, 0x20 );
}



/* Used only to know the size off code NewKeyboardHandler
ALWAYS MUST CAME AFTER NewKeyboardHandler.
This information is used in installation process.
**********************************************************/
void NewKeyboardHandler___ ()
{
}



/* Install keyboard handler if it's not installed
**********************************************************/
void CKeyboard::mInstallKeyboardHandler ()
{
   // verify if is not installed
   if ( ! bINSTALL ) {
      // save old keyboard handler
      _go32_dpmi_get_protected_mode_interrupt_vector ( 9, &stOLD_KEYBOARD_HANDLER );

      // save memory where new keyboard handler is from virtual memory swaps
      _go32_dpmi_lock_code ( (void*) NewKeyboardHandler, (PTRVAL) NewKeyboardHandler___ - (PTRVAL) NewKeyboardHandler );

      // save memory where keys table is from virtual memory swaps
      _go32_dpmi_lock_data ( ayKEYS, sizeof ( ayKEYS ) );

      // substitue current keyboard handler to class keyboard handler
      stNEW_KEYBOARD_HANDLER.pm_offset   = (PTRVAL) NewKeyboardHandler;
      stNEW_KEYBOARD_HANDLER.pm_selector = _go32_my_cs();
      _go32_dpmi_allocate_iret_wrapper ( &stNEW_KEYBOARD_HANDLER );
      _go32_dpmi_set_protected_mode_interrupt_vector ( 9, &stNEW_KEYBOARD_HANDLER );

      // reset key table (all FALSE)
      memset ( (void*) ayKEYS, FALSE, sizeof ( ayKEYS ) );

      // flag handler intalalled
      bINSTALL = m_bInstallHandler = TRUE;
   }
}



/* Desinstall keyboard handler if it's installed. Use
CKeyboard::mInstallKeyboardHandler to reinstall.
**********************************************************/
void CKeyboard::mDesinstallKeyboardHandler ()
{
   // verify if it's installed
   if ( bINSTALL ) {

      // wait all keys release (else you can have GPF)
      mWaitAllRelease ();

      // swap current handler with old
      _go32_dpmi_set_protected_mode_interrupt_vector ( 9, &stOLD_KEYBOARD_HANDLER );

      // free wrapper
      _go32_dpmi_free_iret_wrapper ( &stNEW_KEYBOARD_HANDLER );

      // flag desinstall
      bINSTALL = m_bInstallHandler = FALSE;
   }
}

#endif  // TRG_DOSDJGPP



// Functions implemented only on X Window port
#ifdef TRG_LINUXXWIN

/* Set a specific key in key table acording the given
 key pressed/released in X Window event. This function
is used to translate key codes between X and 3DFC.
dwXKeyCode - the X Window key code
yValue - value to set the key
**********************************************************/
void CKeyboard::mSetKeyTable ( DWORD dwXKeyCode, BYTE yValue )
{
   BYTE
      y3DFCKey = 0;   // receive the 3DFC key code
      
   // make translation
   switch ( dwXKeyCode ) {
   
      case XK_BackSpace:
         y3DFCKey = K_BACKSPACE;
         break;

      case XK_Tab:
         y3DFCKey = K_TAB;
         break;
   
      case XK_Return:
         y3DFCKey = K_ENTER;
         break;
   
      case XK_Escape:
         y3DFCKey = K_ESC;
         break;
   

      case XK_Home:
         y3DFCKey = K_HOME;
         break;
   
      case XK_End:
         y3DFCKey = K_END;
         break;
   
      case XK_Left:
         y3DFCKey = K_LEFT;
         break;
   
      case XK_Right:
         y3DFCKey = K_RIGHT;
         break;
   
      case XK_Up:
         y3DFCKey = K_UP;
         break;
   
      case XK_Down:
         y3DFCKey = K_DOWN;
         break;
   
      case XK_Page_Up:
         y3DFCKey = K_PG_UP;
         break;
   
      case XK_Page_Down:
         y3DFCKey = K_PG_DN;
         break;

   
      case XK_KP_Enter:
         y3DFCKey = K_ENTER + SPECIAL;
         break;

      case XK_KP_Home:
         y3DFCKey = K_HOME + SPECIAL;
         break;
   
      case XK_KP_End:
         y3DFCKey = K_END + SPECIAL;
         break;
   
      case XK_KP_Left:
         y3DFCKey = K_LEFT + SPECIAL;
         break;
   
      case XK_KP_Right:
         y3DFCKey = K_RIGHT + SPECIAL;
         break;
   
      case XK_KP_Up:
         y3DFCKey = K_UP + SPECIAL;
         break;
   
      case XK_KP_Down:
         y3DFCKey = K_DOWN + SPECIAL;
         break;
   
      case XK_KP_Page_Up:
         y3DFCKey = K_PG_UP + SPECIAL;
         break;
   
      case XK_KP_Page_Down:
         y3DFCKey = K_PG_DN + SPECIAL;
         break;
   

      /* REVIEW who are this X keys?
      case XK_KP_0:
         y3DFCKey = K_0 + SPECIAL;
         break;

      case XK_KP_1:
         y3DFCKey = K_1 + SPECIAL;
         break;

      case XK_KP_2:
         y3DFCKey = K_2 + SPECIAL;
         break;

      case XK_KP_3:
         y3DFCKey = K_3 + SPECIAL;
         break;

      case XK_KP_4:
         y3DFCKey = K_4 + SPECIAL;
         break;

      case XK_KP_5:
         y3DFCKey = K_5 + SPECIAL;
         break;

      case XK_KP_6:
         y3DFCKey = K_6 + SPECIAL;
         break;

      case XK_KP_7:
         y3DFCKey = K_7 + SPECIAL;
         break;

      case XK_KP_8:
         y3DFCKey = K_8 + SPECIAL;
         break;

      case XK_KP_9:
         y3DFCKey = K_9 + SPECIAL;
         break;
      */

      // REVIEW - put the macros, not the numbers!
      case 65457:
         y3DFCKey = K_END; // + SPECIAL;
         break;
      
      case 65458:
         y3DFCKey = K_DOWN + SPECIAL;
         break;
      
      case 65459:
         y3DFCKey = K_PG_DN; // + SPECIAL;
         break;
      
      case 65460:
         y3DFCKey = K_LEFT + SPECIAL;
         break;
      
      case 65461:
         y3DFCKey = K_5 + SPECIAL;  // REVIEW who is midle key 5???
         break;
      
      case 65462:
         y3DFCKey = K_RIGHT + SPECIAL;
         break;

      case 65463:
         y3DFCKey = K_HOME; // + SPECIAL;
         break;
      
      case 65464:
         y3DFCKey = K_UP + SPECIAL;
         break;
      
      case 65465:
         y3DFCKey = K_PG_UP; // + SPECIAL;
         break;
      
   
      case XK_F1:
         y3DFCKey = K_F1;
         break;
   
      case XK_F2:
         y3DFCKey = K_F2;
         break;

      case XK_F3:
         y3DFCKey = K_F3;
         break;

      case XK_F4:
         y3DFCKey = K_F4;
         break;

      case XK_F5:
         y3DFCKey = K_F5;
         break;

      case XK_F6:
         y3DFCKey = K_F6;
         break;

      case XK_F7:
         y3DFCKey = K_F7;
         break;

      case XK_F8:
         y3DFCKey = K_F8;
         break;

      case XK_F9:
         y3DFCKey = K_F9;
         break;

      case XK_F10:
         y3DFCKey = K_F10;
         break;

      case XK_F11:
         y3DFCKey = K_F11;
         break;

      case XK_F12:
         y3DFCKey = K_F12;
         break;


      case XK_Shift_L:
         y3DFCKey = K_SHIFT_L;
         break;

      case XK_Shift_R:
         y3DFCKey = K_SHIFT_R;
         break;

      case XK_Control_L:
         y3DFCKey = K_CTRL;
         break;

      case XK_Control_R:
         y3DFCKey = K_CTRL + SPECIAL;
         break;

      case XK_Alt_L:
         y3DFCKey = K_ALT;
         break;
         
      case XK_Alt_R:
         y3DFCKey = K_ALT + SPECIAL;
         break;
         

      // REVIEW - put the macros, not the numbers!
      case 97:
         y3DFCKey = K_A;
         break;

      case 98:
         y3DFCKey = K_B;
         break;

      case 99:
         y3DFCKey = K_C;
         break;

      case 100:
         y3DFCKey = K_D;
         break;

      case 101:
         y3DFCKey = K_E;
         break;

      case 102:
         y3DFCKey = K_F;
         break;

      case 103:
         y3DFCKey = K_G;
         break;

      case 104:
         y3DFCKey = K_K;
         break;

      case 105:
         y3DFCKey = K_I;
         break;

      case 106:
         y3DFCKey = K_J;
         break;

      case 107:
         y3DFCKey = K_K;
         break;

      case 108:
         y3DFCKey = K_L;
         break;

      case 109:
         y3DFCKey = K_M;
         break;

      case 110:
         y3DFCKey = K_N;
         break;

      case 111:
         y3DFCKey = K_O;
         break;

      case 112:
         y3DFCKey = K_P;
         break;

      case 113:
         y3DFCKey = K_Q;
         break;

      case 114:
         y3DFCKey = K_R;
         break;

      case 115:
         y3DFCKey = K_S;
         break;

      case 116:
         y3DFCKey = K_T;
         break;

      case 117:
         y3DFCKey = K_U;
         break;

      case 118:
         y3DFCKey = K_V;
         break;

      case 119:
         y3DFCKey = K_W;
         break;
         
      case 120:
         y3DFCKey = K_X;
         break;

      case 121:
         y3DFCKey = K_Y;
         break;

      case 122:
         y3DFCKey = K_Z;
         break;

      default:        // if key conversion not found
         return;
   }
         
   // set key table
   * ( ayKEYS + y3DFCKey ) = yValue;
}

#endif    // TRG_LINUXXWIN



/* Verify if a specific key is pressed now.
 return TRUE - yes; FALSE - no
**********************************************************/
BOOL CKeyboard::mTestKeyPress ( BYTE yKey )
{
   return * ( ayKEYS + yKey );
}



/* Return the code of the first key pressed found.
If no key pressed return 0.
**********************************************************/
BYTE CKeyboard::mTestAllPress ()
{
   // see all keys's table
   for ( UINT16 ucT = 0; ucT < 256; ucT ++ )
      // if find TRUE return it's code
      if ( mTestKeyPress ( ucT ) == TRUE )
         return ucT;

   return 0;
}



/* Wait some key be pressed. If bTestPress == TRUE then
only wait key pressed after all keys be released.
Default bTestPress = TRUE. return key code.
**********************************************************/
BYTE CKeyboard::mWaitAllPress ( BOOL bTestPress )
{
   // verify if must wait all keys be realeased
   if ( bTestPress ) {
      BOOL
         bRelease;
      // while one key be pressed
      do {
         bRelease = TRUE;
         // see all key's table
         for ( UINT16 ucT = 0; ucT < 256; ucT ++ )
            // if find one pressed
            if ( mTestKeyPress ( ucT ) == TRUE )
               bRelease = FALSE;
      } while ( bRelease == FALSE );
   }

   // while any key not pressed
   while ( 1 )
      // see all key's table
      for ( UINT16 ucT = 0; ucT < 256; ucT ++ )
         // if find one pressed return his code
         if ( mTestKeyPress ( ucT ) == TRUE )
            return ucT;
}



/* Wait all key be released
**********************************************************/
void CKeyboard::mWaitAllRelease ()
{
   BOOL
      bRelease;
   // while any key be pressed
   do {
      bRelease = TRUE;
      // see key's table
      for ( UINT16 ucT = 0; ucT < 256; ucT ++ )
         // if find any key pressed
         if ( mTestKeyPress ( ucT ) == TRUE )
            bRelease = FALSE;
   } while ( bRelease == FALSE );
}
